/******************************************************************************
 *
 *  Filename:       interrupt.c
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 * 25May08 - Changed debounce so entire process uses polling. Original version
 *           was triggered by interrupt for initial press, then used polling to
 *           debounce. Spec sheet advises against mixing interrupts and polling
 *           and the new version works a lot more reliably.
 *
 *
 *****************************************************************************/

/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include "io_cfg.h"             // I/O pin mapping
#include "system/typedefs.h"
#include "user/mxg_error.h"
#include "system/interrupt/interrupt.h"
#include "user/mxg_usart.h"
#include "user/mxg_msgs.h"
#include "user/mxg_iface.h"

/** V A R I A B L E S ********************************************************/
// Timers
volatile byte displayTmr;       // Refresh timer for the LCD
volatile WORD dispTimeout;      // Timeout timer for display
volatile byte rsInjectTmr;      // Timeout timer for injecting running status
                                // ...used in mxg_umsm.c
volatile byte debounceTmr;      // Debounce timer for the buttons
volatile byte swStatus;         // Switch status (for interrupt handler)
volatile byte savedSwStatus;    // ...saved version
volatile BOOL bChange;
// This is just used as a target to receive data from the EUSART if our input 
// buffer is full. Better to make it a global to keep the ISR as fast as we can.
byte readrcvd;

/** INTERRUPT ROUTINES *******************************************************/
#pragma code 

/***********************************************************************
 * Main ISR, invoked from the remapped high prioriy ISR vector
 * (hardware priority not used)
 *
 * Priority implemented by service order:
 * 1. 1ms timer 0 interrupt
 * 2. EUSART receive interrupt
 * 3. EUSART transmit interrupt
 *
 * TODO: I2C interrupts??
 *
 * Calling functions from within the ISR adds quite a bit of code to the
 * context save for the interrupt, so everything is implemented inline.
 **********************************************************************/
void _rm_high_ISR(void)
{
	if(INTCONbits.TMR0IE && INTCONbits.TMR0IF) {
		// Handle timer 0 interrupt
		TMR0L=INTERVAL;				// Restore the timer value
		mClearTimerInt();			// Clear the interrupt flag

		// Now update the system timers by decrementing if not zero
		if(displayTmr) --displayTmr;
        if(dispTimeout._word) --dispTimeout._word;
        if(rsInjectTmr) --rsInjectTmr;

        // Check for user input
        swStatus = PORTB & SWMASK;      // Read switch status
        if (swStatus==savedSwStatus){   // Does it match the saved state?
            debounceTmr=DEBOUNCE;       // ... yes, so init the debounce timer
        }
        else {
            --debounceTmr;              // ... no, decrement the debounce timer
            if(!debounceTmr){           // Have we completed the debounce?
                savedSwStatus = swStatus;   // Yes, save the status
                bChange = TRUE;             // ...indicate keys have changed
                debounceTmr = DEBOUNCE;     // ...reset debounce timer 
            }    
        }
	}

	// Now handle received data from the EUSART
	else if(PIE1bits.RCIE && PIR1bits.RCIF) {
		if(RCSTAbits.OERR==1) {
			// Handle overrun error - bump counter and reset EUSART
            RCSTAbits.CREN=0;
            RCSTAbits.CREN=1;
            sFlags.rcvOERR=ERR_2;
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.rcvOECnt!=0xFFFF)errctrs.rcvOECnt++;
		}
		if(RCSTAbits.FERR==1) {
            // An EUSART framing error has occurred. The data is discarded
            // ...and the error count bumped. The flag is cleared on the next
            // ...good read.
            sFlags.rcvFERR=ERR_1;
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.rcvFECnt!=0xFFFF)errctrs.rcvFECnt++;
		}
		/*****************/
		// GetRcvdByte();
		/*****************/
    	if(mRxBufNotFull()) {
        	*rxbend.bufptr = RCREG;
        	++rxbend.bytes.lo;
        	rxbend.bytes.lo&=rxbMask;
    	}
    	else {                  // The receive buffer is full - ERROR!
        	readrcvd = RCREG;        // Clear the interrupt
        	sFlags.rcvBufFull=ERR_0;
            ERRFLAG = ERROR;                // Show flag on Home display
        	if(errctrs.rcvBFCnt!=0xFFFF)errctrs.rcvBFCnt++;
    	}
		/*****************/
	}
    // ... next in line is the transmit data
	else if(PIE1bits.TXIE && PIR1bits.TXIF) {
        TXREG = *txbstart.bufptr;        // Get the data from the tx buffer
        ++txbstart.bytes.lo;             // Bump the pointer
        txbstart.bytes.lo&=txbMask;      // ... makes sure it wraps properly
        // Disable Tx interrupts if the buffer is empty
        if (mTxBufEmpty()) mDisableTxInt();
    }
}

// Tell the compiler this is our ISR
#pragma interrupt _rm_high_ISR

/** V E C T O R  R E M A P P I N G *******************************************/
extern void _startup (void);        // See c018i.c in your C18 compiler dir

#pragma code _RESET_INTERRUPT_VECTOR = 0x000800
void _reset (void)
{
    _asm goto _startup _endasm
}

#pragma code _HIGH_INTERRUPT_VECTOR = 0x000808
void _high_ISR (void)
{
	_asm goto _rm_high_ISR _endasm
}

/* We're not using priority....
#pragma code _LOW_INTERRUPT_VECTOR = 0x000818
void _low_ISR (void)
{
}
*/

/** I N T E R R U P T  V E C T O R S *****************************************/
// Need this to be able to simulate
#pragma code _REAL_HIGH_INTERRUPT_VECTOR = 0x000008
void _real_high_ISR (void)
{
    _asm goto _high_ISR _endasm
}
#pragma code

/** EOF interrupt.c **********************************************************/
