/*********************************************************************
 *
 *             Microchip USB C18 Firmware -  MIDI Streaming
 *
 *********************************************************************
 * FileName:        usbms.c
 * Dependencies:    See INCLUDES section below
 * Processor:       PIC18
 * Compiler:        C18 2.30.01+
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the Company) for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 * Author               Date        Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Rawin Rojvanit       11/19/04    Original.
 ********************************************************************/

/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include "system\typedefs.h"
#include "system\usb\usb.h"

#ifdef USB_USE_MS

/** V A R I A B L E S ********************************************************/
#pragma udata
byte usbms_rx_len;

/** P R I V A T E  P R O T O T Y P E S ***************************************/

/** D E C L A R A T I O N S **************************************************/
#pragma code

/** U S E R  A P I ***********************************************************/

/******************************************************************************
 * Function:        void USBMSInitEP(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        USBMSInitEP initializes the MS Streaming endpoints, buffer
 *                  descriptors, internal state-machine, and variables.
 *                  It should be called after the USB host has sent out a
 *                  SET_CONFIGURATION request.
 *                  See USBStdSetCfgHandler() in usb9.c for examples.
 *
 * Note:            None
 *****************************************************************************/
void USBMSInitEP(void)
{   
    usbms_rx_len = 0;
    
    USBMS_UEP = EP_OUT_IN|HSHK_EN;             // Enable 2 data pipes

    /*
     * Do not have to init Cnt of IN pipes here.
     * Reason:  Number of bytes to send to the host
     *          varies from one transaction to
     *          another. Cnt should equal the exact
     *          number of bytes to transmit for
     *          a given IN transaction.
     *          This number of bytes will only
     *          be known right before the data is
     *          sent.
     */
    USBMS_BD_OUT.Cnt = sizeof(usbms_out);     // Set buffer size
    USBMS_BD_OUT.ADR = (byte*)&usbms_out;     // Set buffer address
    USBMS_BD_OUT.Stat._byte = _USIE|_DAT0|_DTSEN;// Set status

    USBMS_BD_IN.ADR = (byte*)&usbms_in;      // Set buffer address
    USBMS_BD_IN.Stat._byte = _UCPU|_DAT1;      // Set buffer status

}//end USBMSInitEP

/******************************************************************************
 * Function:        void USBMSWrite(byte *buffer, byte len)
 *
 * PreCondition:    mUSBMSTxIsBusy() must return false.
 *
 *                  Value of 'len' must be equal to or smaller than
 *                  USBMS_EP_SIZE
 *                  For an interrupt/bulk endpoint, the largest buffer size is
 *                  64 bytes.
 *
 * Input:           buffer  : Pointer to the starting location of data bytes
 *                  len     : Number of bytes to be transferred
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Use this macro to transfer data located in data memory.
 *
 *                  Remember: mUSBMSTxIsBusy() must return false before user
 *                  can call this function.
 *                  Unexpected behavior will occur if this function is called
 *                  when mUSBMSTxIsBusy() != 0
 *
 *                  Typical Usage:
 *                  if(!mUSBMSTxIsBusy())
 *                      USBMSWrite(buffer, 3);
 *
 * Note:            None
 *****************************************************************************/
void USBMSWrite(byte *buffer, byte len)
{
	byte i;
	
    /*
     * Value of len should be equal to or smaller than USBMS_EP_SIZE.
     * This check forces the value of len to meet the precondition.
     */
	if(len > USBMS_EP_SIZE)
	    len = USBMS_EP_SIZE;

   /*
    * Copy data from user's buffer to dual-ram buffer
    */
    for (i = 0; i < len; i++)
    	usbms_in[i] = buffer[i];

    USBMS_BD_IN.Cnt = len;
    mUSBBufferReady(USBMS_BD_IN);

}//end USBMSWrite

/******************************************************************************
 * Function:        byte USBMSRead(byte *buffer, byte len)
 *
 * PreCondition:    Value of input argument 'len' should be smaller than the
 *                  maximum endpoint size responsible for receiving report
 *                  data from USB host for HID class.
 *                  Input argument 'buffer' should point to a buffer area that
 *                  is bigger or equal to the size specified by 'len'.
 *
 * Input:           buffer  : Pointer to where received bytes are to be stored
 *                  len     : The number of bytes expected.
 *
 * Output:          The number of bytes copied to buffer.
 *
 * Side Effects:    Publicly accessible variable usbms_rx_len is updated
 *                  with the number of bytes copied to buffer.
 *                  Once USBMSRead is called, subsequent retrieval of
 *                  usbms_rx_len can be done by calling macro
 *                  mUSBMSGetRxLength().
 *
 * Overview:        USBMSRead copies a string of bytes received through
 *                  the OUT endpoint to a user's specified location. 
 *                  It is a non-blocking function. It does not wait
 *                  for data if there is no data available. Instead it returns
 *                  '0' to notify the caller that there is no data available.
 *
 * Note:            If the actual number of bytes received is larger than the
 *                  number of bytes expected (len), only the expected number
 *                  of bytes specified will be copied to buffer.
 *                  If the actual number of bytes received is smaller than the
 *                  number of bytes expected (len), only the actual number
 *                  of bytes received will be copied to buffer.
 *****************************************************************************/
byte USBMSRead(byte *buffer, byte len)
{
    usbms_rx_len = 0;
    
    if(!mUSBMSRxIsBusy())
    {
        /*
         * Adjust the expected number of bytes to equal
         * the actual number of bytes received.
         */
        if(len > USBMS_BD_OUT.Cnt)
            len = USBMS_BD_OUT.Cnt;
        
        /*
         * Copy data from dual-ram buffer to user's buffer
         */
        for(usbms_rx_len = 0; usbms_rx_len < len; usbms_rx_len++)
            buffer[usbms_rx_len] = usbms_out[usbms_rx_len];

        /*
         * Prepare dual-ram buffer for next OUT transaction
         */
        USBMS_BD_OUT.Cnt = sizeof(usbms_out);
        mUSBBufferReady(USBMS_BD_OUT);
    }//end if
    
    return usbms_rx_len;
    
}//end USBMSRead

#endif //def USB_USE_MS

/** EOF usbms.c *************************************************************/
