/******************************************************************************
 *
 *  Filename:       mxg_iface.c
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 * 16Jul08 - changed display flags to support USB connected and single flag for
 *           WX5, MIDI and USB
 * 06Aug08 - Added MIDI IN and MIDI OUT monitoring to the Home display 
 * 08Sep08 - Added reset all error counters code (press Set while at an error
 *           counter that is already zero).
 * 10Sep08 - Added code to clear the global error flag when the last error
 *           counter is cleared.
 * 17Sep08 - Changed interface so that dERRORS is only shown if errors have
 *           occurred.
 * 18Sep08 - Fixed bug in dERRORS display logic and change logic for dSPLASH
 *           doSet() so we go to dERRORS if there's an error.
 *           Also improved usability by only displaying non-zero error counters
 *           unless already in counter screens and all have been reset
 *           individually.
 * 23Sep08 - Added backlight control logic
 * 25Sep08 - Refactored nextChar(). Save and restore brightness from EEPROM
 * 15Oct08 - Switched dBACKLIGHT and dERRORS order so backlight can be reached
 *           easily if an error has occurred.
 * 12Nov08 - Changed the backlight PWM frequency by using a prescale factor of
 *           4 insteand of 16. Frequency is now 23.4kHz instead of 5.8kHz, 
 *           which means it isn't audible if it gets into the audio circuits. 
 *
 *****************************************************************************/
 /******************************************************************************
 * mistralXG user interface handling
 *
 * User interface logic
 * 
 * The display shows a number of standard screens, listed below:
 *
 * dSPLASH      The power on display (Name, firmware version, copyright)
 * dHOME        The "normal" display (Name, user option flags, error flags)
 * dWX5EXT      Selects the external MIDI source (WX5/MIDI IN)
 * dUSBEXT	    Selects USB (if enabled) or WX5/MIDI IN
 * dMIDITHRU    Enables or disables MIDI THRU from WX5/Ext input (see dWX5EXT)
 * dMIDIOUT     Enables or disables MIDI OUT from USB
 * dRUNSTAT     Enables or disables running status for USB MIDI OUT
 * dFILTER      Enables or disables filtering of bad MIDI data
 * dERRORS      Select to view the error counters
 * dBACKLIGHT   Allows the user to control the backlight brightness
 *
 * NOTES:
 *
 * There are two user buttons. bSELECT cycles through the standard screens
 * above, wrapping back to dSPLASH after dBACKLIGHT. dERRORS is only displayed
 * if errors have occurred. All screens revert to dHOME after DISP_TIMEOUT
 * seconds. Each screen that controls a user setting reacts accordingly if bSET
 * is pressed. That is, each press of bSET toggles the setting controlled by
 * the particular screen, or bumps the backlight brightness, wrapping to OFF
 * after maximum.
 *
 * Pressing bSET:
 * - at dHOME displays dBACKLIGHT, for easy adjustment of the LCD brightness,
 *   or dERRORS if an error has occurred.
 *
 * - while at dERRORS takes the user into a cycle of error count
 *   screens, one for each type of error that is tracked by the code. Pressing
 *   bSET again at any error screen resets the related counter if not zero and
 *   allows the user to reset all error counters if zero. bSELECT cycles
 *   through the error counts. The error screens also honour the
 *   DISP_TIMEOUT second timeout and return to dHOME.
 * 
 * This is implemented as yet another state machine, with the following state
 * transition table:
 *
 * ---------------------------------------------------------------------------
 * Current display | bSelect            | bSet                | n-sec timeout
 * ---------------------------------------------------------------------------
 * dSPLASH         | dHOME              | dHOME               | dHOME
 * dHOME           | dWX5EXT            | dSPLASH             | -
 * dWX5EXT         | dUSBEXT            | Toggle select       | dHOME
 * dUSBEXT	       | dMIDITHRU          | Toggle select       | dHOME
 * dMIDITHRU       | dMIDIOUT           | Toggle enable       | dHOME
 * dMIDIOUT        | dRUNSTAT           | Toggle enable       | dHOME
 * dRUNSTAT        | dFILTER            | Toggle enable       | dHOME
 * dFILTER         | dERRORS(1)         | Toggle enable       | dHOME
 * dERRORS(1)      | dBACKLIGHT         | First error counter | dHOME
 * dBACKLIGHT      | dSPLASH            | Change brightness   | dHOME
 *                 |                    |                     |
 * Error counter   | Next error counter | Reset counter (if!0)| dHOME
 *                 |                    | Offers to reset all |
 *                 |                    | counters (if 0)     |
 * dRESETALL       | Back to error ctr  | Reset all counters  | dHOME
 * ---------------------------------------------------------------------------
 * (1) Note that dERRORS is only displayed if an error has occurred.
 *
 * BE CAREFUL: Non-portable methods have been used in some places for reasons
 *             of efficiency. This has been noted in particular cases.
 *
 *****************************************************************************/
#include <p18cxxx.h>
#include "io_cfg.h"             // I/O pin mapping
#include "system\typedefs.h"
#include "system\interrupt\interrupt.h"
#include "user\mxg_msgs.h"
#include "user\mxg_xlcd.h"
#include "user\mxg_error.h"
#include "user\mxg_iface.h"
#include "user\mxg_eeprom.h"
#include "user\user.h"

// Some local variables for the non-blocking LCD update routines
screen nextScreen, savedScreen;
sPTR line0, line1, tmpptr;

// Keeps track of how many more chars to be written to the display
byte WritesToLCDPending;
// Keeps track of current display brightness
byte brightness;
BOOL brightChange;              // Indicates that the brightness has changed

// We don't want the trailing zero, so truncate
rom char hexchar[16] = "0123456789ABCDEF";

/** INTERNAL FUNCTIONS ***/
void setDispPtrs(void);
void nextIntensity(void);


/** Initialise the interface **/
void initInterface(void){
    byte i;

    // We don't have to worry about interrupts as they're not enabled yet
    setDisplay(dSPLASH);
    debounceTmr=DEBOUNCE;               // Init switch debounce timer
    savedSwStatus=swStatus=bNone;       // Init switch state to none pressed
    bChange=FALSE;                      // ... and no change
    for(i=0;i<16;i++){
        sActivity[i]._byte = NO_MIDI;

    // Set the initial display brightness to the value stored in EEPROM
    // Only 9 values are valid 0xn0, where n=0-8. To ensure we don't pick up
    // bad data the very first time the EEPROM is read, we mask the value
    brightness = readEEPROM(SAVED_BRITE);
    brightness &= 0xF0;
    if(brightness>0x80) brightness = 0x80;
    brightChange = FALSE;

#ifndef CHK_TIMING /* If not defined, set up PWM on RC1 - see io_cfg.h */

    // Set the period by setting PR2
    // PWM Period = [(PR2) + 1]  4  TOSC  (TMR2 Prescale Value)
    // Slowest PWM freq is PR2=255 => period=256*4*(1/48E6)*16=0.341ms or 2930Hz
    // We're using PR2=0x7F, prescale=4 => 128*4*(1/48E6)*4 = 42.67us/23.44kHz
    PR2 = 0x7F;
    // Set the duty cycle = (CCPR2L:CCP2CON<5:4>)  TOSC  (TMR2 Prescale Value)
    // If the PWM duty cycle value is longer than the PWM period, the CCPx
    // pin will not be cleared. If duty cycle=0, the CCPx pin will not be set.
    CCPR2L = brightness;
    // RC1 is already set to output by mInitIO()
    // Setup T2CON 
    // 0-unimplemented, 0000-postscale (n/a), 1-TMR2 enable, 01- 4 prescale
    T2CON = 0b00000101;
    // Setup CCP2CON
    // 00 - unimplemented, 00 - duty cycle LSB[1:0], 11xx - PWM mode
    CCP2CON = 0b00001100;

#endif /* CHK_TIMING */

    }
}

/** Routines to handle button presses */
// First the Select button
void doSelect(void){
    byte i;

    // Let's take a look at the "normal" screens first...just cycle round them
    if(nextScreen>=dSPLASH){            // Are we in "normal" screens?
        if(nextScreen<dBACKLIGHT) {     // Yes... do we need to wrap?
            nextScreen++;               // No.. bump to next screen
            // Check if we're at the Error display and do we need to show it?
            if(nextScreen==dERRORS && ERRFLAG!=ERROR){
                // We're at Error display but no errors so skip it
                nextScreen++;
            }
            setDisplay(nextScreen); 
        }
        else {                          
            // Yes, we're at dBACKLIGHT, so go to dSPLASH
            setDisplay(dSPLASH);   // ... otherwise, wrap to splash screen
        }
    }
    else{           // We're in the error screens, so cycle round them.
        if(nextScreen==dRESETALL){      // Check if we were in Reset all state
            // Return to the error screen that got us here
            setDisplay(savedScreen);
        }
        else {                          // We're in the basic error screens
            // Test each successive flag, starting at the next one, wrapping
            // round and back to the current one and then one more if no
            // counters are set. The effect is that we go to the next non-zero
            // counter or, if all counters are zero, step through one-by-one.
            for (i=0;i<=16;i++) {
                ++nextScreen;
                nextScreen &=0x0f;
                if(sFlags.fa[nextScreen]!=NOERR) break;
            }
            setDisplay(nextScreen);
        }
    }
}

// Now the Set button
void doSet(void){
    byte i;

    switch(nextScreen){
        case dHOME:
        case dSPLASH:
            if(ERRFLAG==ERROR)              // If errors have occurred
                setDisplay(dERRORS);        // ... show the error screen
            else
                setDisplay(dBACKLIGHT);     // ... otherwise the backlight
            break;
        case dWX5EXT:
          	WX5Sel = !WX5Sel;       // TRUE: WX5 data, FALSE: Ext MIDI IN data
            optsChange = TRUE;
            if(WX5Sel) {
                bWX5Sel = 1;
            }
            else {
                bWX5Sel = 0;
            }
            break;
        case dUSBEXT:
	        EnIntMIDI = !EnIntMIDI; // TRUE:PC USB, FALSE: Data from WX5/Ext
            optsChange = TRUE;
            if(EnIntMIDI){
                bIntMIDISel = 1;
            }
            else{
                bIntMIDISel = 0;
            }
            break;
        case dMIDITHRU:
	        EnMIDIThru = !EnMIDIThru; // TRUE:PC USB, FALSE: Data from WX5/Ext
            optsChange = TRUE;
            if(EnMIDIThru){
                THRUFLAG = EnThruTRUE;
                bEnMIDIThru = 1;
            }
            else{
                THRUFLAG = EnThruFALSE;
                bEnMIDIThru = 0;
            }
            break;
        case dMIDIOUT:
	        EnMIDIOut = !EnMIDIOut; // TRUE:PC USB, FALSE: Data from WX5/Ext
            optsChange = TRUE;
            if(EnMIDIOut){
                OUTFLAG = EnOutTRUE;
                bEnMIDIOut = 1;
            }
            else{
                OUTFLAG = EnOutFALSE;
                bEnMIDIOut = 0;
            }
            break;
        case dRUNSTAT:
    	    EnRunStatus = !EnRunStatus;
            optsChange = TRUE;
            RSFLAG = EnRunStatus?EnRSTRUE:EnRSFALSE;
            break;
        case dFILTER:
            FilterMIDIIn = !FilterMIDIIn;
            optsChange = TRUE;
            FILTFLAG = FilterMIDIIn?EnFiltTRUE:EnFiltFALSE;
            break;
        case dERRORS:
            // Moving to error screens... display the first non-zero counter.
            // We should only get here if there is a non-zero counter.
            tmpptr.sRam = sFlags.fa;
            for (i=0;i<16;i++) {
                if(*tmpptr.sRam!=NOERR) break;
                tmpptr.sRam++;
            }
            setDisplay(dERR0+i);
            break;
        case dBACKLIGHT:
            nextIntensity();
            break;
        case dRESETALL:
            clearErrors();
            setDisplay(dHOME);
            break;
        // The default section handles Set for error screens
        default:
            // Pressing Set on the error screens resets the current counter
            // if it is not zero, otherwise the user is offered the chance to
            // reset all of the counters
            if (errctrs.ca[nextScreen] != 0) {

                // Used to test if all errors are now clear
                char testflags;
                
                // The EUSART error counters are volatile (used in interrupts)
                // so disable interrupts while we do the reset. Changing the 
                // displayed flag has to be atomic with resetting the counter.
                mDisableRxInt();
                errctrs.ca[nextScreen] = 0;         // Clear the counter
                sFlags.fa[nextScreen]=NOERR;        // Clear the error flag
                mEnableRxInt();

                // OK - now let's test to see if we've cleared the last error
                // Could use a loop but unrolled to avoid indexing and ensure
                // top speed. We test the flags that are not set in the 
                // interrupt routine last and with interrupts off as these can
                // change without us knowing.
                // NON-PORTABLE: NOERR = 0x5F - ANDing any of the error flags
                // used with it will change the value. If different encoding
                // or flag values are used, this may fail!!
                testflags = sFlags.DATNotExpctd;
                testflags &= sFlags.VM1NotExpctd;
                testflags &= sFlags.VM2NotExpctd;
                testflags &= sFlags.SC0NotExpctd;
                testflags &= sFlags.SC1NotExpctd;
                testflags &= sFlags.SC2NotExpctd;
                testflags &= sFlags.SYXNotExpctd;
                testflags &= sFlags.EOXNotExpctd;
                testflags &= sFlags.SC0F4Rcvd;
                testflags &= sFlags.SC0F5Rcvd;
                testflags &= sFlags.SRTF9Rcvd;
                testflags &= sFlags.SRTFDRcvd;
                testflags &= sFlags.MSBOFlow;

                // Now we check the volatile flags with EUSART interrupts off
                // We also clear the global flag on the home display while
                // interrupts are off so that a EUSART error can't sneak in
                // while we're doing it.
                mDisableRxInt();
                testflags &= sFlags.rcvBufFull;
                testflags &= sFlags.rcvFERR;
                testflags &= sFlags.rcvOERR;
                // If all errors are clear, clear the global error flag
                if(testflags==NOERR) ERRFLAG = NOERRS;
                mEnableRxInt();
            }
            else {
                // Save the current error screen so we can return
                savedScreen = nextScreen;
                setDisplay(dRESETALL);
            }
            break;
    }
    // Set the timeout to return to the home display
    // We have to disable timer interrupts temporarily as we are setting a 
    // word value and this is not atomic
    mDisableTimerInt();
    dispTimeout._word=DISP_TIMEOUT;         // Display timeout period
    mEnableTimerInt();
}


/********************************************************************
* Function Name:nextIntensity()
* Return Value:	void
* Parameters:	void
* Description:	Bumps the intensity of the backlight, wraps back to
*               being OFF when max is reached.
********************************************************************/
void nextIntensity(void){
    if (brightness==0x80) brightness = 0;   // Wrap back to zero at max
    else brightness += 0x10;                // ...else next value
    brightChange = TRUE;
    
#ifndef CHK_TIMING /* If not defined, adjust the display brightness */
    CCPR2L = brightness;
#endif /* CHK_TIMING */
}

/********************************************************************
* Function Name:setDisplay(screen scrn)
* Return Value:	void
* Parameters:	scrn: the screen to be displayed
* Description:	This routine sets up a string to be written to the Hitachi
*               HD44780 LCD controller. It cues it up to be written by the main
*               loop in user.c, which writes one character at a time by calling
*               nextChar(). This is done to avoid having to block while the LCD
*               is busy and thus keeps the USB going. 
********************************************************************/
void setDisplay(screen scrn){
    nextScreen=scrn;        // Save the screen to be displayed
    setDispPtrs();          // Set the pointers to update the display
    // Now set the counter to initiate the next update
    WritesToLCDPending=34;  // 34 = (Move cursor + 16 chars)*2

    // Set the timeout to return to the home display
    // We have to disable timer interrupts temporarily as we are setting a 
    // word value and this is not atomic
    mDisableTimerInt();
    dispTimeout._word=DISP_TIMEOUT;         // Display timeout period
    mEnableTimerInt();
}

void refreshDisplay(void) {
    WritesToLCDPending=34;  // 34 = (Move cursor + 16 chars)*2
    setDispPtrs();
}

void setDispPtrs(void){
    switch(nextScreen){
        case dHOME:
            line0.sRam = sHome;
            line1.sRam1 = sActivity;
            break;
        case dWX5EXT:
            line0.sRom = sExtMidi;
            line1.sRom = sExtInput;
            break;
        case dUSBEXT:
            line0.sRom = sExtInt;
            line1.sRom = sSynthSel;
            break;
        case dMIDITHRU:
            line0.sRom = sExtThru;
            line1.sRom = sEnable;
            break;
        case dMIDIOUT:
            line0.sRom = sMidiOut;
            line1.sRom = sEnable;
            break;
        case dRUNSTAT:
            line0.sRom = sRunStat;
            line1.sRom = sEnable;
            break;
        case dFILTER:
            line0.sRom = sFilter;
            line1.sRom = sEnable;
            break;
        case dERRORS:
            line0.sRom = sErrors;
            line1.sRam = sFlags.fa;
            break;
        case dBACKLIGHT:
            line0.sRom = sBacklight;
            // line1 not used as we create the display on the fly
            break;
        case dSPLASH:
            line0.sRom = sSplash;
            line1.sRom = sCpyrght;
            break;
        case dRESETALL:
            line0.sRom = eResetCtrs;
            line1.sRom = eResetAll;
            break;    
        // Default sets up pointers for error counter screens
        default:
            line0.sRom = errdisp[nextScreen];
            line1.sRom = eResetErr;
            break;
    }
    displayTmr=DISP_REFRESH;
}

/********************************************************************
 * Function Name:nextChar(void)
 * Return Value:	void
 * Parameters:	
 * Description:	This routine writes the next character from the buffers set up
 *              by setDisplay to the LCD. The data is written to the character
 *              generator RAM or the display data RAM depending on which
 *              SetxxRamAddr routine was called previously. 
 *
 ********************************************************************/
void nextChar(void)
{
    // Keeps track of where we are drawing the brightness indicator bar
    char briteindx;
    char nextch;                        // The next char to go to the display.

    if(WritesToLCDPending==34) {        // Starting a new screen, home cursor
        SetDDRamAddr(STARTROW1);
    } 
    else if(WritesToLCDPending==17) {   // Starting second line, position cursor
        SetDDRamAddr(STARTROW2);
        // brightness is one of 0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70,
        // 0x80. Shift right 3 bits turns this into a number from 0-16 which we
        // can use as an index for the brightness indicator if we're on that
        // screen
        briteindx=brightness>>3;
    }
    else if(WritesToLCDPending>17){     // First line, write it out
        if(nextScreen==dHOME){          // dHOME is RAM-based
            nextch = *line0.sRam++;
        }
        else {                          // Other screens are ROM-based
            nextch = *line0.sRom++; 
        }
            WriteDataXLCD(nextch);
    }
    else{       // WritesToLCDPending < 17, write second line
        switch(nextScreen){
            case dHOME:
                // Should be *line1.sRam1 really, but WriteDataXLCD wants a 
                // char argument and sRam is compatible with sRam1.
                // Write the current char from sActivity, which shows MIDI
                // activity since the last refresh
                nextch = *line1.sRam;
                // Reset the activity character and bump the pointer.
                *line1.sRam++ = NO_MIDI;
                break;
            case dBACKLIGHT:
                // If first char, write "unselected" char
                if(WritesToLCDPending==16) nextch = UNSELECT;
                // ...otherwise write the indicator or a space, depending on
                // current brightness indicator count
                else nextch = briteindx?INDICATOR:' ';
                // Decrement the brightness indicator count if not zero
                if(briteindx) briteindx--;
                break;
            case dERRORS:
                nextch = *line1.sRam++; 
                break;
            case dWX5EXT:
                if(WritesToLCDPending==16) {
                    nextch = WX5Sel?SELECT:UNSELECT; 
                    *line1.sRom++;
                }
                else if(WritesToLCDPending==8){
                    nextch = !WX5Sel?SELECT:UNSELECT;
                    *line1.sRom++;
                }
                else nextch = *line1.sRom++; 
                break;
            case dUSBEXT:
                if(WritesToLCDPending==16) {
                    nextch = !EnIntMIDI?SELECT:UNSELECT; 
                    *line1.sRom++;
                }
                else if(WritesToLCDPending==8){
                    nextch = EnIntMIDI?SELECT:UNSELECT;
                    *line1.sRom++;
                }
                else nextch = *line1.sRom++; 
                break;
            case dMIDITHRU:
                if(WritesToLCDPending==16) {
                    nextch = EnMIDIThru?SELECT:UNSELECT; 
                    *line1.sRom++;
                }
                else if(WritesToLCDPending==8){
                    nextch = !EnMIDIThru?SELECT:UNSELECT;
                    *line1.sRom++;
                }
                else nextch = *line1.sRom++; 
                break;
            case dMIDIOUT:
                if(WritesToLCDPending==16) {
                    nextch = EnMIDIOut?SELECT:UNSELECT; 
                    *line1.sRom++;
                }
                else if(WritesToLCDPending==8){
                    nextch = !EnMIDIOut?SELECT:UNSELECT;
                    *line1.sRom++;
                }
                else nextch = *line1.sRom++; 
                break;
            case dRUNSTAT:
                if(WritesToLCDPending==16) {
                    nextch = EnRunStatus?SELECT:UNSELECT; 
                    *line1.sRom++;
                }
                else if(WritesToLCDPending==8){
                    nextch = !EnRunStatus?SELECT:UNSELECT;
                    *line1.sRom++;
                }
                else nextch = *line1.sRom++; 
                break;
            case dFILTER:
                if(WritesToLCDPending==16) {
                    nextch = FilterMIDIIn?SELECT:UNSELECT; 
                    *line1.sRom++;
                }
                else if(WritesToLCDPending==8){
                    nextch = !FilterMIDIIn?SELECT:UNSELECT;
                    *line1.sRom++;
                }
                else nextch = *line1.sRom++; 
                break;
            case dSPLASH:
            case dRESETALL:
                nextch = *line1.sRom++; 
                break;
            // Default writes the 2nd line of the error count display
            default:
                switch(WritesToLCDPending){
                    // Last 4 characters are the error counter
                    case 4:
                        nextch = hexchar[errctrs.ca[nextScreen]>>12&0x0F];
                        break;
                    case 3:
                        nextch = hexchar[errctrs.ca[nextScreen]>>8&0x0F];
                        break;
                    case 2:
                        nextch = hexchar[errctrs.ca[nextScreen]>>4&0x0F];
                        break;
                    case 1:
                        nextch = hexchar[errctrs.ca[nextScreen]&0x0F];
                        break;
                    // The first 12 characters are normal text
                    default:
                        nextch = *line1.sRom++;
                        break;
                }
                break;
        }
        WriteDataXLCD(nextch);
    }
    WritesToLCDPending--;
}

