/******************************************************************************
 *
 *  Filename:       mxg_musm.c
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 * 06Aug08 - Added MIDI IN and MIDI OUT monitoring to the Home display
 * 09Aug08 - Use USBActive to prevent MIDI being sent when not connected
 *
 *****************************************************************************/

/******************************************************************************
 * MIDI->USB-MIDI STATE MACHINE
 *****************************************************************************/
#include <p18cxxx.h>
#include "user\user.h"
#include "user\mxg_usart.h"     // Include before mxg_musm.h
#include "user\mxg_musm.h"
#include "user\mxg_umsm.h"
#include "user\mxg_msgs.h"
#include "user\mxg_iface.h"
#include "user/mxg_error.h"
#include "system\interrupt\interrupt.h"

/****************************************************************************
 * This module takes an input MIDI data stream from the EUSART and converts
 * it into USB-MIDI packets for sending to the PC. The conversion is 
 * performed by a finite state machine that keeps track of the MIDI status
 * as the stream is processed to form the correct USB-MIDI packets. Its
 * operation is documented in the state table below.
 *
 * It's worth noting that the design allows for any malformed packets to 
 * be transferred across the interface to the PC as is (within the constraints
 * of USB-MIDI). A flag, FilterMIDIIN, is set to TRUE by default and this 
 * removes invalid packets from the input stream (with the exception of
 * incorrectly terminated SYSEX streams, the first parts of which will have
 * been passed to the PC before the invalid terminator is received). This is
 * in line with the MIDI spec. If, for some reason, it is necessary for the 
 * malformed data to reach the PC, this flag should be set to FALSE.
 *
 * State transition table to turn incoming MIDI stream into USB-MIDI data
 *
 * Current| Byte received
 * State  | DAT	   VM1    VM2    SC0    SC1    SC2    SYX     EOX    SRT
 * -----------------------------------------------------------------------
 * S0     | S0???  N1S1   N2S2   S0*    N1S0   N2S0   SYSEX   S0?    S0**
 * N1S0   | S0*    N1S1?? N2S2?? S0?    N1S0?? N2S0?? SYSEX?? S0?    N1S0**
 * N2S0   | N1S0   N1S1?? N2S2?? S0?    N1S0?? N2S0?? SYSEX?? S0?    N2S0**
 * S1     | S1*!   N1S1   N2S2   S0*    N1S0   N2S0   SYSEX   S0?    S1**
 * N1S1   | S1*    N1S1?? N2S2?? S0?    N1S0?? N2S0?? SYSEX?? S0?    N1S1**
 * S2     | N1S2!  N1S1   N2S2   S0*    N1S0   N2S0   SYSEX   S0?    S2**
 * N1S2   | S2*    N1S1?? N2S2?? S0?    N1S0?? N2S0?? SYSEX?? S0?    N1S2**
 * N2S2   | N1S2   N1S1?? N2S2?? S0?    N1S0?? N2S0?? SYSEX?? S0?    N2S2**
 * SYSEX  | SYSEX* N1S1?? N2S2?? S0?    N1S0?? N2S0?? SYSEX?? S0*    SYSEX**
 *
 *
 * Categories for bytes received (0xnn) - undefined codes
 *
 * DAT 0x00-0x7f       SC0 (0xF4,0xF5),0xF6   SYX 0xF0
 * VM1 0xCn,0xDn       SC1 0xF1,0xF3          EOX 0XF7
 * VM2 0x8n-0xBn,0xEn  SC2 0xF2               SRT 0xF8-0xFF (0xF9,0xFD)
 * 
 *                            
 * Normal operation:
 *
 * !  Running status
 * *  Output an assembled USB-MIDI packet. Sysex has to assemble and output
 *    packets until EOX received
 * ** Output a System Realtime packet, maintain current packet assembly status
 *
 *
 * Invalid data sequence operation:
 *
 * ?   Handle partial MIDI-USB packet already assembled and output the single
 *     byte status
 * ??  Handle partial MIDI-USB packet already assembled and change status,
 *     starting to assemble new MIDI-USB packet
 * ??? Output single data byte MIDI-USB packet
 *
 *
 * Status |Expected byte  |Comment
 * ---------------------------------------------------------------------------
 * S0     |Status         |MIDI message complete, waiting for next status byte,
 *        |               |running status invalid
 * N1S0   |Data           |One data byte required to complete MIDI message,
 *        |               |previous status does not support running status
 * N2S0   |Data           |Two data bytes required to complete MIDI message,
 *        |               |previous status does not support running status
 * S1     |Status or data |MIDI message complete, waiting for next status byte,
 *        |               |running status valid (one data byte required)
 * N1S1   |Data           |One data byte required to complete MIDI message,
 *        |               |previous status supports running status (which
 *        |               |needs one data byte)
 * S2     |Status or data |MIDI message complete, waiting for next status byte,
 *        |               |running status valid (two data bytes required)
 * N1S2   |Data           |One data byte required to complete MIDI message,
 *        |               |previous status supports running status (which
 *        |               |needs two data bytes)
 * N2S2   |Data           |Two data bytes required to complete MIDI message,
 *        |               |previous status supports running status (which
 *        |               |needs two data bytes)
 * SYSEX  |Data or EOX    |Data is part of the Sysex message, EOX terminates
 *        |               |this status, no running status
 ******************************************************************************/

/* Private Function Prototypes *********/
void sendRealtime(void);
void sendPacket(void); 
void sh_invalid(void);
void sh_newStatus(void);
void outputPartialPkt(void);

// Handlers for the possible states of the state machine
near rom void sh_s0(void);
near rom void sh_n1s0(void);
near rom void sh_n2s0(void);
near rom void sh_s1(void);
near rom void sh_n1s1(void);
near rom void sh_s2(void);
near rom void sh_n1s2(void);
near rom void sh_n2s2(void);
near rom void sh_sysex(void);

// Enum for the state machine states
enum msState {S0, N1S0, N2S0, S1, N1S1, S2, N1S2, N2S2, SYSEX};
typedef enum msState msState;
// Enum for the MIDI byte types
enum byteType {DAT, VM1, VM2, SC0, SC1, SC2, SYX, EOX, SRT};
typedef enum byteType byteType;

// Array of pointers to functions to handle each of the state machine's states
// - the order is critical and must match the msState enum in mxg_musm.h.
near rom void (* rom handlerptr[])(void) 
				= {sh_s0, sh_n1s0, sh_n2s0, 
				   sh_s1, sh_n1s1, 
				   sh_s2, sh_n1s2, sh_n2s2, 
				   sh_sysex};

// Array of system status byte types (0xF0-0xF7)
rom byteType statusTypes[]
            = {SYX,SC1,SC2,SC1,SC0,SC0,SC0,EOX};

/********************************************************************
 * MIDI status data
 * The USB-MIDI packet is 4-bytes long, the first byte is in two nibbles:
 * <cable>:<code index>. In this application, the cable for the input stream
 * is always 0. The code index varies based on the input data... see the USB
 * MIDI spec for details.
 ********************************************************************/
msState CurrentState;
byteType ByteType;
byte  	RunningStatus;  // Keeps track of the status for running status

union {                 // The USB-MIDI packet being assembled
    byte a[4];          // Defined as a union for faster access
    struct {
        byte b0;
        byte b1;
        byte b2;
        byte b3;
    } bytes;
} MIDIUSBPacket;
byte    packetBytes;    // Keeps track of how much data is in the packet
//byte	MIDICable;      // Always cable 0 for this app.
byte    NextByte;
BOOL    SysexActive;
bp   UMIptr;            // Pointer to the next free packet space in usbmidiin

// We queue outgoing packets here... in most cases, only single packets
// are likely to be sent to the USB SIE.
// IMPORTANT: This needs to be at address 0x??c0 so that the low byte is 0 
// if the buffer is full. NOTE: This is dependent on USBMS_EP_SIZE being 0x40!
#pragma udata usbmidiinbuf=0x06c0
byte	USBMidiIn[USBMS_EP_SIZE];
#pragma udata

/********************************************************************
 * Function Name:	initMusm()
 * Return Value:	void
 * Parameters:		void
 * Description:    	This initialises the state machine ready to start 
 *					processing MIDI data from MIDI devices to the PC.
 ********************************************************************/
void initMusm(void)
{
	CurrentState = S0;
	//MIDICable = 0x00;
    packetBytes=0;
	RunningStatus=0;
    SysexActive=FALSE;
    UMIptr.bufptr=USBMidiIn;
}

/********************************************************************
 * Function Name:	getByteType()
 * Return Value:	void
 * Parameters:		void
 * Description:    	This routine returns the kind of byte that has been
 *                  received in the MIDI stream so that it can be
 *                  handled appropriately.
 ********************************************************************/
byteType getByteType(void)
{
    if(NextByte<0x80) return DAT;   // 0x00-0x7F are data bytes
    if(NextByte>=0xF8) return SRT;  // 0xF8-0xFF are System Realtime
    if(NextByte<0xC0) return VM2;   // 0x80-0xBF are Voice messages with 2 data
    if(NextByte<0xE0) return VM1;   // 0xC0-0xDF are voice messages with 1 data
    if(NextByte<0xF0) return VM2;   // 0xE0-0xEF are Voice messages with 2 data
    else return statusTypes[NextByte&0x0F];
}

/********************************************************************
 * Function Name:	processNextByte()
 * Return Value:	void
 * Parameters:		void
 * Description:    	This checks for data in the input buffer and processes
 *					found data, using it to build a USB-MIDI packet. When
 *					the packet is complete, it is send... we always send
 *                  4-byte packets so that the host PC gets any data as soon
 *                  as possible after we've received it.
 *
 *                  We have to disable interrupts as we are manipulating 
 *                  pointers that are used in the ISR and thus they may
 *                  change unexpectedly if we don't.
 ********************************************************************/
void processNextByte(void) 
{
	// Get the next byte from the receive buffer
	NextByte = *rxbstart.bufptr;
	// ...and bump the receive buffer start pointer
	mDisableRxInt();
	++rxbstart.bytes.lo;
	rxbstart.bytes.lo&=rxbMask;
	mEnableRxInt();

	ByteType = getByteType();
    // Check to see if we've just received a System Realtime status byte
    // If so, we can send it straight away as SRT bytes can be sent at any
    // time without affecting MIDI status
	if(ByteType==SRT) {
		sendRealtime();
	}
	else {
		(*handlerptr[CurrentState])();
	}
}

/********************************************************************
 * Function Name:	sh_<state>()
 * Return Value:	void
 * Parameters:		void
 * Description:    	These routines implement the state machine for 
 *                  converting the incoming MIDI data stream into properly
 *                  formed USB-MIDI packets. When one of these routines is
 *                  invoked, NextByte contains the next byte from the 
 *                  incoming stream, and ByteType contains its type. These
 *                  two data and the CurrentState determine what happens 
 *                  and the next state of the state machine (see table above).
 *
 *                  SRT (System Realtime) status bytes have already been 
 *                  dealt with so do not need to be handled here. 
 *********************************************************************
 * Function Name:	sh_s0()
 *
 * MIDI message complete, waiting for next status byte running status invalid
 ********************************************************************/
void sh_s0(void)
{
    // This is a rest point for the input MIDI stream, so we are expecting a 
    // new status byte. S0 is the only state for which receiving DAT is invalid
    // so we handle this locally.
    if(ByteType==DAT){
        sFlags.DATNotExpctd=ERR_3;      // ... bump error count and stay in S0
        ERRFLAG = ERROR;                // Show flag on Home display
        if(errctrs.DATNECnt!=0xFFFF)errctrs.DATNECnt++; 
	    // Send single data byte
        MIDIUSBPacket.bytes.b1=NextByte;
        packetBytes=2; 
        outputPartialPkt(); 
    }
    // ...otherwise, we're changing status
    else sh_newStatus();    
}

/********************************************************************
 * Function Name:	sh_n1s0()
 *
 * One data byte required to complete MIDI message, previous status does not 
 * support running status
 ********************************************************************/
void sh_n1s0(void)
{
    if(ByteType==DAT){
        MIDIUSBPacket.a[packetBytes++]=NextByte;
        sendPacket();           // OK, we're done, send the packet
        CurrentState=S0;
    }
    else sh_invalid();
}

/********************************************************************
 * Function Name:	sh_n2s0()
 *
 * Two data bytes required to complete MIDI message, previous status does not
 * support running status
 ********************************************************************/
void sh_n2s0(void)
{
    if(ByteType==DAT){
        MIDIUSBPacket.a[packetBytes++]=NextByte;
        CurrentState=N1S0;
    }
    else sh_invalid();
}

/********************************************************************
 * Function Name:	sh_s1()
 *
 * MIDI message complete, waiting for next status byte, running status valid
 * (one data byte required)
 ********************************************************************/
void sh_s1(void)
{
    // This is a rest point for the input MIDI stream, so we are expecting a 
    // new status byte. The one exception is running status, where we have to
    // supply the status byte and the received byte is the one data byte 
    // required to complete the packet. We set the first byte based on the
    // status byte's type and send the packet.
    if(ByteType==DAT){ 
        // Running status, send packet and stay in S1
        MIDIUSBPacket.bytes.b0=RunningStatus>>4;
        MIDIUSBPacket.bytes.b1=RunningStatus;
        MIDIUSBPacket.bytes.b2=NextByte;
        packetBytes=3;
        sendPacket();
    }
    else sh_newStatus();    // ...otherwise, we're changing status
}

/********************************************************************
 * Function Name:	sh_n1s1()
 *
 * One data byte required to complete MIDI message, previous status supports
 * running status (which needs one data byte)
 ********************************************************************/
void sh_n1s1(void)
{
    if(ByteType==DAT){
        MIDIUSBPacket.a[packetBytes++]=NextByte;
        sendPacket();
        CurrentState=S1;
    }
    else sh_invalid();
}

/********************************************************************
 * Function Name:	sh_s2()
 *
 * MIDI message complete, waiting for next status byte, running status valid
 * (two data bytes required)
 ********************************************************************/
void sh_s2(void)
{
    // This is a rest point for the input MIDI stream, so we are expecting a 
    // new status byte. The one exception is running status, where we have to
    // supply the status byte and the received byte is the first of the two
    // data bytes required to complete the packet. We set the first byte based
    // on the status byte's type and set status to N1S2 to wait for the final
    // data byte.
    if(ByteType==DAT){
    // Running status,start new packet, move to N1S2
        MIDIUSBPacket.bytes.b0=RunningStatus>>4;
        MIDIUSBPacket.bytes.b1=RunningStatus;
        MIDIUSBPacket.bytes.b2=NextByte;
        packetBytes=3;
        CurrentState=N1S2;
    }
    else sh_newStatus();    // ...otherwise, we're changing status
}
            
/********************************************************************
 * Function Name:	sh_n1s2()
 *
 * One data byte required to complete MIDI message, previous status supports
 * running status (which needs two data bytes)
 ********************************************************************/
void sh_n1s2(void)
{
    if(ByteType==DAT){
        MIDIUSBPacket.a[packetBytes++]=NextByte;
        sendPacket();
        CurrentState=S2;
    }
    else sh_invalid();
}

/********************************************************************
 * Function Name:	sh_n2s2()
 *
 * Two data bytes required to complete MIDI message, previous status supports
 * running status (which needs two data bytes)
 ********************************************************************/
void sh_n2s2(void)
{
    if(ByteType==DAT){
        MIDIUSBPacket.a[packetBytes++]=NextByte;
        CurrentState=N1S2;
    }
    else sh_invalid();
}

/********************************************************************
 * Function Name:	sh_sysex()
 *
 * Data is part of the Sysex message, EOX terminates this status,
 * no running status
 ********************************************************************/
void sh_sysex(void)
{

    switch(ByteType){
        case DAT:
            if(packetBytes==0){     
                // Starting a new packet with continuing
                // Sysex data. Always cable 0 for this app.
                MIDIUSBPacket.bytes.b0=0x04; 
                packetBytes=1;
            }
            // Add the incoming byte to the packet
            MIDIUSBPacket.a[packetBytes++]=NextByte;
            if(packetBytes==4){
                // We've filled this packet, so send it, and get ready for 
                // the next byte by resetting the byte counter.
                sendPacket();
                packetBytes=0;
            }
            break;
        case EOX:
            if (packetBytes==1){
                MIDIUSBPacket.bytes.b0=0x05;
                MIDIUSBPacket.bytes.b2=0x00;
                MIDIUSBPacket.bytes.b3=0x00;
            }
            else if (packetBytes==2){
                MIDIUSBPacket.bytes.b0=0x06;
                MIDIUSBPacket.bytes.b3=0x00;
            }
            else if (packetBytes==3){
                MIDIUSBPacket.bytes.b0=0x07;
            }
            MIDIUSBPacket.a[packetBytes++]=0xF7;
            sendPacket();
            CurrentState=S0;
            break;
        default:
            sh_invalid();
            break;
    }
}

/********************************************************************
 * Function Name:	sh_newStatus()
 *
 * Handles new incoming status bytes for S0, S1 and S2. EOX is invalid
 ********************************************************************/            
void sh_newStatus(void) {
    // First we initialise the new USB-MIDI packet byte putting the status
    // byte in byte 1 and clearing bytes 2 and 3
    MIDIUSBPacket.bytes.b1=NextByte;
    MIDIUSBPacket.bytes.b2=0;
    MIDIUSBPacket.bytes.b3=0;
    // We'll have two bytes in the packet when sh_newStatus() is done
    packetBytes=2;   
    // Next, set the USB-MIDI <cable>:CIN byte      
    switch(ByteType){            
        case VM1:
            MIDIUSBPacket.bytes.b0=NextByte>>4;
            CurrentState=N1S1;
            RunningStatus=NextByte;     // Save running status
            break;
        case VM2:
            MIDIUSBPacket.bytes.b0=NextByte>>4;
            CurrentState=N2S2;
            RunningStatus=NextByte;     // Save running status
            break;
        case SC0:
            MIDIUSBPacket.bytes.b0=0x05;
            // SC0 is 0xF6, 0xF5 or 0xF4. Only 0xF6 is valid so we filter
            // 0xF4 and 0xF5 unless FilterMIDIIN==FALSE (default is TRUE)
            if((NextByte&0x02)==0) {            // Test for 0xF4 or 0xF5
                if(NextByte==0xF4) {            // 0xF4
                    sFlags.SC0F4Rcvd=ERR_B;     // Bump error count
                    ERRFLAG = ERROR;            // Show flag on Home display
                    if(errctrs.SC0F4Cnt!=0xFFFF)errctrs.SC0F4Cnt++; 
                }
                else {                          // 0xF5
                    sFlags.SC0F5Rcvd=ERR_C;     // Bump error count
                    ERRFLAG = ERROR;            // Show flag on Home display
                    if(errctrs.SC0F5Cnt!=0xFFFF)errctrs.SC0F5Cnt++; 
                }
            }
            if(NextByte==0xF6 || !FilterMIDIIn){
                sendPacket();           // Send the single byte and move to S0
            }
            CurrentState=S0;
            RunningStatus=0;
            break;
        case SC1:
            MIDIUSBPacket.bytes.b0=0x02;
            CurrentState=N1S0;
            RunningStatus=0;
            break;
        case SC2:
            MIDIUSBPacket.bytes.b0=0x03;
            CurrentState=N2S0;
            RunningStatus=0;
            break;
        case SYX:
            // With sysex messages, we don't know what the first byte of 
            // the packet will be until later, but in most cases it will
            // be <cable>:4
            MIDIUSBPacket.bytes.b0=0x04;// Always cable zero for this app.
            SysexActive=TRUE;
            CurrentState=SYSEX;
            RunningStatus=0;
            break;
        case EOX:
            // EOX is invalid unless we're in a Sysex transfer. This means that
            // we have to handle EOX slightly differently. All other invalid
            // status bytes arrive when not in S0, S1 or S2 and are dealt with
            // in sh_invalid(), which calls sh_newStatus() to establish the
            //  new status. Because EOX is also invalid in the three "rest"
            // states, sh_newStatus() is not called by sh_invalid() for EOX.
            // This means we have to manage the error counter and partial packet
            // output here.
            sFlags.EOXNotExpctd=ERR_A;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.EOXNECnt!=0xFFFF)errctrs.EOXNECnt++;  
            outputPartialPkt();             // Output partial packet 
            CurrentState=S0;         
            break;
        default:
            break;
    }  
}

/********************************************************************
 * Function Name:	sh_invalid()
 *
 * Handles incoming status bytes when not in S0, S1 or S2. That is, 
 * we're expecting a data byte but a status byte turns up. The MIDI spec
 * says that in this situation, we have to adopt the new status and continue.
 * This code gives the user the option to discard malformed packets or to
 * pass them on to the PC (using the single byte USB-MIDI packet).
 ********************************************************************/  
void sh_invalid(void)
{
    RunningStatus=0;                // Clear any running status
    SysexActive=FALSE;              // Terminate Sysex status
    // Output the partially formed packet that has been interrupted
    outputPartialPkt();             

    switch(ByteType){
        case VM1:
            sFlags.VM1NotExpctd=ERR_4;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.VM1NECnt!=0xFFFF)errctrs.VM1NECnt++;             
            break;
        case VM2:
            sFlags.VM2NotExpctd=ERR_5;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.VM2NECnt!=0xFFFF)errctrs.VM2NECnt++;             
            break;
        case SC0:
            sFlags.SC0NotExpctd=ERR_6;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.SC0NECnt!=0xFFFF)errctrs.SC0NECnt++;             
            break;
        case SC1:
            sFlags.SC1NotExpctd=ERR_7;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.SC1NECnt!=0xFFFF)errctrs.SC1NECnt++;             
            break;
        case SC2:
            sFlags.SC2NotExpctd=ERR_8;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.SC2NECnt!=0xFFFF)errctrs.SC2NECnt++;             
            break;
        case SYX:
            sFlags.SYXNotExpctd=ERR_9;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.SYXNECnt!=0xFFFF)errctrs.SYXNECnt++;             
            SysexActive=TRUE;
            break;
        case EOX:
            sFlags.EOXNotExpctd=ERR_A;      // Bump error count
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.EOXNECnt!=0xFFFF)errctrs.EOXNECnt++; 
	        // Send single EOX byte. 
            // CurrentState=S0;packetBytes=0; set by outputPartialPkt()
            MIDIUSBPacket.bytes.b1=NextByte;
            packetBytes=2; 
            outputPartialPkt(); 
            break;
        default:
            // Should never get here!
            break;
    }
    if(ByteType!=EOX) sh_newStatus();  // Status change
}

/********************************************************************
 * Function Name:	sendRealtime()
 * Return Value:	void
 * Parameters:		void
 * Description:    	When this routine is called, NextByte contains a 
 *                  System Realtime (SRT) status byte. These can be present
 *                  at any point in a MIDI data stream and do not affect any
 *                  other aspects (such as running status). An SRT USB-MIDI
 *                  packet is assembled and passed to the PC via USB.
 *                  If FilterMIDIIN is TRUE (default), then the invalid bytes
 *                  0xF9 and 0xFD are not sent.
 ********************************************************************/
void sendRealtime(void)
{
    // Check for 0xF9 or 0xFD - these are invalid according to the MIDI spec
    if(NextByte==0xF9){
        sFlags.SRTF9Rcvd=ERR_D;     // Bump error count
        ERRFLAG = ERROR;            // Show flag on Home display
        if(errctrs.SRTF9Cnt!=0xFFFF)errctrs.SRTF9Cnt++; 
    }
    if(NextByte==0xFD) {
        sFlags.SRTFDRcvd=ERR_E;     // Bump error count
        ERRFLAG = ERROR;            // Show flag on Home display
        if(errctrs.SRTFDCnt!=0xFFFF)errctrs.SRTFDCnt++; 
    }
    // Now output the byte if USB is active, unless it's invalid and we're
    // filtering the stream. Because we're only looking at values between 0xF8
    // and 0xFF, we can check for 0xF9 and 0xFD by testing just the lower 2 bits
    if(USBActive && (((NextByte&0x03)!=0x01) || !FilterMIDIIn)){
        // Do this inline for speed
        *UMIptr.bufptr=0x0f;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=NextByte;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=0;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=0;
        UMIptr.bytes.lo++;
        if(!UMIptr.bytes.lo) {
            sFlags.MSBOFlow=ERR_F;          // MIDI staging buffer overflow
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.MSBOFCnt!=0xFFFF)errctrs.MSBOFCnt++; // Bump error count
            UMIptr.bytes.lo-=4;             // ... discard previous packet 
        }
    }
}

/********************************************************************
 * Function Name:	outputPartialPkt()
 * Return Value:	void
 * Parameters:		0<=packetBytes<4 is set to the packet byte count. The count
 *                  will never be 4 because then a correctly-formed packet will
 *                  have been assembled and sent by sendPacket().
 *                  If we choose to send the individual bytes, only bytes 2 and
 *                  3 of the packet are sent (if present) as byte 0 is the
 *                  USB-MIDI <cable>:<CIN> byte.
 * Description:    	When this routine is called, MIDIUSBPacket usually contains
 *                  a malformed packet due to the out-of-sequence arrival of
 *                  a status byte (or a data byte in state S0). This gives us
 *                  the option of disassembling the partial packet and sending
 *                  the individual bytes in a sequence of USB-MIDI single byte
 *                  (CIN=0xf) packets. The MIDI spec says that invalid data is
 *                  to be discarded, so the default is for FilterMIDIIN to be
 *                  true. For Sysex, we may have sent several valid Sysex stream
 *                  packets before the invalid state occurs, we cannot recall
 *                  them, so we just avoid sending the last partial packet.
 *                  This routine can be called with packetBytes==0 when a 
 *                  full Sysex packet has just been sent and the next byte
 *                  received is neither a DAT or EOX byte.
 ********************************************************************/
void outputPartialPkt(void)
{
    if(USBActive && (!FilterMIDIIn && packetBytes)) {
        // At this point, we'll either have one or two MIDI bytes assembled in
        // the MIDIUSBPacket (and the USB-MIDI cable/CIN byte, which we don't
        // need). So we pick them up and output as single-byte packets
        // Do this inline for speed
        *UMIptr.bufptr=0x0f;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=MIDIUSBPacket.a[1];
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=0;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=0;
        UMIptr.bytes.lo++;
        if(!UMIptr.bytes.lo) {
            sFlags.MSBOFlow=ERR_F;          // MIDI staging buffer overflow
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.MSBOFCnt!=0xFFFF)errctrs.MSBOFCnt++; // Bump error count
            UMIptr.bytes.lo-=4;             // ... discard previous packet 
        }
        if(packetBytes==3){
            *UMIptr.bufptr=0x0f;
            UMIptr.bytes.lo++;
            *UMIptr.bufptr=MIDIUSBPacket.a[2];
            UMIptr.bytes.lo++;
            *UMIptr.bufptr=0;
            UMIptr.bytes.lo++;
            *UMIptr.bufptr=0;
            UMIptr.bytes.lo++;
            if(!UMIptr.bytes.lo) {
                sFlags.MSBOFlow=ERR_F;          // MIDI staging buffer overflow
                ERRFLAG = ERROR;                // Show flag on Home display
                if(errctrs.MSBOFCnt!=0xFFFF)errctrs.MSBOFCnt++; // Bump count
                UMIptr.bytes.lo-=4;             // ... discard previous packet 
            }   
        }
    }
    packetBytes=0;
    CurrentState=S0;
}

/********************************************************************
 * Function Name:	sendPacket()
 * Return Value:	void
 * Parameters:		void
 * Description:    	When this routine is called, MIDIUSBPacket contains a 
 *                  correctly formed USB-MIDI packet ready to be transmitted
 *                  to the PC via USB. We place the packet in the queue ready
 *                  to be sent to the PC.
 ********************************************************************/
void sendPacket(void)
{
    // Check if channel message 0x8n-0xEn and update MIDI activity if so
    if(MIDIUSBPacket.bytes.b1>=0x80 && MIDIUSBPacket.bytes.b1<0xF0){
        sActivity[MIDIUSBPacket.bytes.b1&0x0f].b0 = 1;
    }
    // If USB is active, queue up the packet for sending
    if(USBActive) {
        // Do this inline for speed
        *UMIptr.bufptr=MIDIUSBPacket.bytes.b0;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=MIDIUSBPacket.bytes.b1;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=MIDIUSBPacket.bytes.b2;
        UMIptr.bytes.lo++;
        *UMIptr.bufptr=MIDIUSBPacket.bytes.b3;
        UMIptr.bytes.lo++;
        if(!UMIptr.bytes.lo) {
            sFlags.MSBOFlow=ERR_F;          // MIDI staging buffer overflow
            ERRFLAG = ERROR;                // Show flag on Home display
            if(errctrs.MSBOFCnt!=0xFFFF)errctrs.MSBOFCnt++; // Bump error count
            UMIptr.bytes.lo-=4;             // ... discard previous packet 
        }
    }
    packetBytes=0;
}