/******************************************************************************
 *
 *  Filename:       mxg_umsm.c
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 * 07Jun08 - Implemented status injection after a fixed period with a constant
 *           running status (rsTmr) as suggested by Thorsten Klose.
 * 06Aug08 - Added MIDI IN and MIDI OUT monitoring to the Home display
 *
 *****************************************************************************/
 
/******************************************************************************
 * mistralXG USB-MIDI>MIDI routines 
 *****************************************************************************/
#include <p18cxxx.h>
#include "user\user.h"
#include "user\mxg_usart.h"     // Needed for the buffer pointer definition
#include "user\mxg_umsm.h"
#include "user\mxg_iface.h"
#include "user\mxg_msgs.h"
#include "system\usb\class\ms\usbms.h"
#include "system\interrupt\interrupt.h"


bp   UMOptr;            // Pointer to the next free packet space in usbmidiout
// We queue outgoing packets in USBMidiOut. In most cases, only single packets
// are likely to be sent to the USB SIE.
// IMPORTANT: This needs to be at address 0x??80 so that the low byte pointer 
// arithmetic works.
#pragma udata usbmidioutbuf=0x0680
byte	USBMidiOut[USBMS_EP_SIZE];
#pragma udata

// Array containing the data byte count for the 16 USB-MIDI CIN values
rom byte packetSize[] = {0,0,2,3,3,1,2,3,3,3,3,3,2,2,3,1};

// Keep track of the running status in the output stream if the user
// has selected to enable it.
byte RunningStatusOut;

/********************************************************************
 * Function Name:	initUmsm()
 * Return Value:	void
 * Parameters:		void
 * Description:    	This initialises the state machine ready to start 
 *					processing MIDI data from the PC to MIDI devices.
 ********************************************************************/
void initUmsm(void)
{
    UMOptr.bufptr=USBMidiOut;   // Set the pointer for the output buffer
    RunningStatusOut=0;         // Clear running status
    rsInjectTmr=RS_TIMEOUT;     // Set running status timeout (mxg_umsm.c)
}

/********************************************************************
 * Function Name:	processNextBuffer
 * Return Value:	void
 * Parameters:
 * Description:		USBMidiOut has been loaded with the latest transfer
 *                  from the PC USB. This routine works its way through
 *                  the buffer, transferring the MIDI data to the EUSART
 *                  transmit buffer. If selected by the user, Running status
 *                  is utilised to minimise the amount of MIDI data sent.
 *
 *                  Note that this routine is only called when we know there
 *                  is enough room in the output Tx buffer, so we don't need
 *                  to check here.
 ********************************************************************/
void processNextBuffer(void)
{
    byte len, cable, count, firstByte;
    BOOL channelMsg;

    // Set the pointer to the start of the buffer
    UMOptr.bytes.lo=0x80;
    
    // Get the length of data in the buffer and divide by 4 to get the number
    // of USB-MIDI packets, then iterate until the number of packets remaining
    // is zero.
    for(len=mUSBMSGetRxLength()>>2;len;len--){
//!!    // We currently have only one embedded MIDI IN Jack, so cable will
        // always be zero. This may change when the memory for patches is added.
        cable=*UMOptr.bufptr>>4;
        // Get the number of data bytes in this packet
        count=packetSize[(*UMOptr.bufptr)&0x0f];
        // We've got the info we need from the <cable>:<CIN> byte so bump ptr.
        ++UMOptr.bytes.lo;
        // See if we need to update the MIDI activity monitor
        channelMsg = FALSE;             // Assume it's not a channel message
        firstByte=*UMOptr.bufptr;       // Get the packet's 1st MIDI byte
        if(firstByte>=0x80 && firstByte<0xF0){ // Channel message 0x8n-0xEn
            channelMsg = TRUE;
            sActivity[firstByte&0x0f].b1 = 1;
        }
        // Deal with running status if enabled
        if(EnRunStatus && count){
            // Check if it is a MIDI status byte that supports running status
            if(channelMsg){
                // Check if the status byte is the same as last time
                // If so, skip the status byte unless the injection timeout
                // has expired, in which case we'll send it anyway so that
                // if something is hot-plugged, it'll get a status byte sooner
                // rather than later (thanks to Thorsten Klose for the idea).
                // As rsInjectTmr is a byte, we're not worried about interrupts
                if (firstByte==RunningStatusOut && rsInjectTmr) {
                    ++UMOptr.bytes.lo;      // Yes, so skip the status byte
                    --count;                // ...and reduce the count by 1
                }
                else {
                    RunningStatusOut=firstByte; // No, set new running status
                    rsInjectTmr=RS_TIMEOUT;     // ...and reset injection timer
                }
            }
            else {                      // Running status not supported
                RunningStatusOut=0;     // ...so clear it
            }
        }

        // Now we move the MIDI data into the tranmit buffer
        while(count--) {
            *txbend.bufptr=*UMOptr.bufptr;
            ++UMOptr.bytes.lo;

            // Bump the transmit buffer end pointer
            mDisableTxInt();    // Disable Tx interrupts in case we're Txing
            ++txbend.bytes.lo;
            txbend.bytes.lo&=txbMask;
            mEnableTxInt();     // Enable Tx interrupt so we can start sending
        }
        // UMOptr is now between 1 and 4 more than when we entered the for loop
        // and we need to make it +4
        UMOptr.bytes.lo+=3;         // Now it's +4 to +7
        UMOptr.bytes.lo&=0xFC;      // Clear the last two bits, so it's +4  
    }   // For loop
}