/******************************************************************************
 *
 *  Filename:       mxg_usart.c
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 *
 *
 *****************************************************************************/
 
/******************************************************************************
 * mistralXG EUSART routines
 *****************************************************************************/
#include <p18cxxx.h>
#include "system\typedefs.h"
#include "system/interrupt/interrupt.h"
#include "user\mxg_usart.h"

/********************************************************************
 * Put the USART buffers in their own section to ensure that they
 * start on an address boundary that allows simpler pointer arithmetic
 * by using masks. Each buffer should be 2^n in size and on a 2^(n+1) boundary.
 * 
 * The receive buffer shouldn't need to be large... MIDI data arrives at
 * 32us/bit or 320us for a byte (with start/stop bits) - we should be able
 * to process that without problems.
 *
 * Data from the USB port comes in up to 64 byte chunks... so we'll allow
 * space for another chunk to arrive while the first is being processed.
 ********************************************************************/
#pragma udata usart_buffers=0x700
volatile byte	txbuffer[TXB_SIZE];
volatile byte	rxbuffer[RXB_SIZE];
#pragma udata
// The pointers to the buffers can go whereever the compiler puts them
volatile bp	txbstart, txbend;
volatile bp	rxbstart, rxbend;

/********************************************************************
 * Function Name:	initUSART
 * Return Value:	void
 * Parameters:		-
 * Description:    	This routine sets up the USART for MIDI.
 ********************************************************************/
void initUSART(void)
{
	// 8-bit, Async, Tx enabled, hi-speed
	TXSTA = USART_ASYNC & USART_8_BIT &	USART_TX_ENABLE & USART_HI_SPEED;	
	// 8-bit, Rx enabled, serial comms enabled
	RCSTA = USART_SPEN & USART_8_BIT & USART_RX_ENABLE; //0x90;						

	// Microchip libary sets TRISC6 to 0. Datasheet has conflicting info:
	// ...Section 20 says <7><6>=0b11, Table 10-5 says <7><6>=0b10
	// ...PDF presentation from web says 0b11
	TRISCbits.TRISC6 = 0;				// Set port pins for EUSART operation
	TRISCbits.TRISC7 = 1;
	// No auto baud, Tx/Rx data not inverted, 8-bit BRG, no Wakeup
	BAUDCON = 0;

	// Set bit rate. 31250bps for MIDI, 38400bps for serial (useful for debug)
    // n=48000000/(16*bitrate) -1 => 95 for 31250bps, 77 for 38400bps
	SPBRG = 95;

	// Initialise receive and transmit buffers. Disable interrupts before doing
    // so as initUSART is called when interrupts are enabled. 
    mDisableTxInt();
    mDisableRxInt();
	rxbstart.bufptr=rxbend.bufptr=rxbuffer;
	txbstart.bufptr=txbend.bufptr=txbuffer;

    // Reset the EUSART to clear any status from data received while we're
    // not accepting interrupts (USB inactive, for example).
    RCSTAbits.CREN=0;
    RCSTAbits.CREN=1;

	// Init interrupt enable bits (GIE/PEIE set in userinit() - see user.c
	mEnableRxInt(); 			// Enable Rx and wait for data
}
