/******************************************************************************
 *
 *  Filename:       mxg_xlcd.c
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 * 19May08 - Modified reset sequence to match Hitachi spec sheet as MCLR reset 
 *           wasn't consistent. Seems to work now.
 *
 *****************************************************************************/
 
/******************************************************************************
 * mistralXG LCD utility routines (adapted from the Microchip library to avoid
 * blocking while writing to the LCD).
 *****************************************************************************/
#include <p18cxxx.h>
#include "system\typedefs.h"
#include "system\interrupt\interrupt.h"
#include "user\mxg_msgs.h"
#include "user\mxg_xlcd.h"


/********************************************************************
 * Function Name:	DelayForLCDEnable
 * Return Value:	void
 * Parameters:		-
 * Description:    	This routine provides a blocking delay to ensure 
 *					that the setup and hold time for the Enable pin are met.
 ********************************************************************/
void DelayForLCDEnable(void)
{
    // Just the call and return is sufficient delay
}

/********************************************************************
 * Function Name:	Delay_x1ms
 * Return Value:	void
 * Parameters:		delay - required delay in milliseconds + 1
 * Description:    	This routine provides a blocking delay for the 
 *					specified time by making use of the 1ms interrupt timer.
 *					Add one to the count used by the timer as there is
 *					a latency of one 1ms timer tick. Timer variables are
 *					defined in interrupt.c with the interrupt routines.					
 ********************************************************************/
void Delay_x1ms(byte delay)
{
	// No need to disable timer interrupts while we set the value as displayTmr
    // is a byte and thus the assignment is atomic
	displayTmr=delay;
	// Checking displayTmr is also atomic
	while(displayTmr);
}

// Define characters for the application specific symbols.
// (Copyright is not always present in the supplied font)
// These are often contained in strings, where it is necessary to use the 
// hex form (\x???) so symbolic equivalents are not defined for all characters.
const rom char charpatt[]={	0b11111,        // Copyright (code 0x08)
						 	0b10001,
						 	0b10101,
						 	0b10111,
						 	0b10101,
						 	0b10001,
						 	0b11111,
							0b00000,

							0b00000,        // Selected (code 0x09)
						 	0b01110,
						 	0b11111,
						 	0b11111,
						 	0b11111,
						 	0b01110,
							0b00000,
							0b00000,

	                        0b00000,        // Not selected (code 0x0A)
						 	0b01110,
						 	0b10001,
						 	0b10001,
						 	0b10001,
						 	0b01110,
							0b00000,
							0b00000,

							0b00100,        // USB enabled (code 0x0B)
						 	0b00101,
						 	0b10101,
						 	0b10110,
						 	0b01100,
						 	0b00100,
							0b00100,
							0b00000,

							0b00000,        // No MIDI activity (code 0x0C)
						 	0b00000,
						 	0b00000,
						 	0b11111,
						 	0b00000,
						 	0b00000,
							0b00000,
							0b11111,

							0b00000,        // Input MIDI activity (code 0x0D)
						 	0b11111,
						 	0b11111,
						 	0b11111,
						 	0b00000,
						 	0b00000,
							0b00000,
							0b11111,

							0b00000,        // Output MIDI activity (code 0x0E)
						 	0b00000,
						 	0b00000,
						 	0b11111,
						 	0b00000,
						 	0b11111,
							0b11111,
							0b11111,

							0b00000,        // Input and output MIDI activity 
						 	0b11111,        // (code 0x0F)
						 	0b11111,
						 	0b11111,
							0b00000,
						 	0b11111,
						 	0b11111,
							0b11111};

/********************************************************************
 * Function Name:	OpenXLCD
 * Return Value:	void
 * Parameters:		lcdtype: sets the type of LCD (lines)
 * Description:    	This routine configures the LCD. Based on the Hitachi 
 * 		  			HD44780 LCD controller. The routine configures the I/O pins
 * 		  			of the microcontroller, sets up the LCD for 4-bit mode and
 * 		  			clears the display. Initialisation needs three delay 
 * 		  			routines:
 *                 	DelayPORXLCD() provides at least 15ms delay
 *                 	DelayXLCD() provides at least 5ms delay
 *                 	DelayFor18TCY() provides a 18 Tcy delay
 *
 * Blocking is allowed in this routine as it is only called when the USB is
 * inactive (so we can use while(BusyXLCD());).
 *
 * Timing notes (for LCD running at 5V - slower timings for low voltage):
 * 
 * Address (RW, RS) set up and hold times to ^E or vE are less than one
 * instruction cycle, even with a 48MHz clock, so no concerns there.
 *
 * Data delay time (from ^E) is 160ns so two instruction cycles with a 
 * 48MHz clock are sufficient to read data back.
 * 
 * mPulseXLCDEnable ensures E is high for at least 230ns. The overall E cycle
 * time must be no less than 500ns. Delaying with E low for four instruction
 * cycles will achieve this with a 48MHz clock. See mxg_xlcd.h for the macro def.
 *
 * The data bits are on the lower 4 bits of the port. The default state is to
 * have them as inputs. BusyXLCD() then doesn't have to change them. All write
 * commands have to set them to outputs and restore them to inputs when done.
 ********************************************************************/
void OpenXLCD(unsigned char lcdtype)
{
		unsigned char i;				// General purpose counter

		// All control signals made outputs
		// RW = 0 for write, 1 for read
		// RS = 0 for command, 1 for data
		// E = 1 to enable
        TRIS_RW = 0;                    
        TRIS_RS = 0;
        TRIS_E = 0;
        RW_PIN = 0;                     // R/W pin made low (0 for write)
        RS_PIN = 0;                     // Register select pin made low
        E_PIN = 0;                      // Clock pin made low

        // Delay for >15ms to allow for LCD Power on reset. Make it 40ms as 
        // this is the value when running the LCD at low voltage (useful when
        // running off 4.5V battery power) and it's a one-time delay
        Delay_x1ms(40);
        
        // Setup interface to LCD
        TRIS_DATA_PORT &= 0xf0;			// Make the data lines outputs
        DATA_PORT &= 0xf0;
        DATA_PORT |= 0b00000011;        // Function set cmd(4-bit interface)

		mPulseXLCDEnable();				// Clock the command in (8-bit at POR)
        
        // Delay for at least 4.1ms (don't need E low delay)
        Delay_x1ms(6);

        // Setup interface to LCD
        DATA_PORT &= 0xf0;              // Function set cmd(4-bit interface)
        DATA_PORT |= 0b00000011;		// Still 8-bit at the moment

		mPulseXLCDEnable();				// Clock the cmd in 
										// (don't need E low delay)
        // Delay for at least 100us - we'll make it at least 1ms
        Delay_x1ms(2);

        // Setup interface to LCD
        DATA_PORT &= 0xf0;              // Function set cmd(4-bit interface)
        DATA_PORT |= 0b00000011;		// Still 8-bit at the moment

		mPulseXLCDEnable();				// Clock the cmd in 

        TRIS_DATA_PORT |= 0x0f;         // Make data nibble input
		
		// At this point, we can start testing for LCD busy
        // Set data interface width, # lines, font
        while(BusyXLCD());              // Wait if LCD busy
        DATA_PORT |= 0b00000010;		// Still 8-bit at the moment
		mPulseXLCDEnable();				// Clock the cmd in 

        while(BusyXLCD());              // Wait if LCD busy
        WriteCmdXLCD(lcdtype);          // Function set cmd

        // Turn the display off
        while(BusyXLCD());
        WriteCmdXLCD(DOFF&CURSOR_OFF&BLINK_OFF); 

		// Define the copyright symbol, since it's not available on all
		// versions of the LCD module
		while(BusyXLCD());
		SetCGRamAddr(0);				// 1st CGRAM char
		for (i=0;i<sizeof(charpatt);i++) {
			while(BusyXLCD());
			WriteDataXLCD(charpatt[i]);
		}

        // Clear display - sets DD Ram address to 0 and increment mode
        while(BusyXLCD());              // Wait if LCD busy
        WriteCmdXLCD(LCD_CLEAR);		// Clear display

        while(BusyXLCD());              // Wait if LCD busy
        WriteCmdXLCD(DON&CURSOR_OFF&BLINK_OFF);     // Display ON, no cursor

        return;
}

/********************************************************************
* Function Name:  BusyXLCD
* Return Value:   char: busy status of LCD controller
* Parameters:     void
* Description:    This routine reads the busy status of the
*                 Hitachi HD44780 LCD controller.
*
* Assumes data bits are set to inputs before being called
********************************************************************/
BOOL BusyXLCD(void)
{
//!! return FALSE;
        RW_PIN = 1;                     // Set the control bits for read
        RS_PIN = 0;
        E_PIN = 1;                      // Clock in the command
		// Data takes 160ns to appear from ^E, so two instructions is enough
		// but we have to ensure E is high long enough, so use
		// DelayforLCDEnable().
        DelayForLCDEnable();				
		// Test for busy bit on lower nibble (4-bit interface)
        // Keep these separate for performance
        if(DATA_PORT&0x08)
        {
                E_PIN = 0;              // Reset clock line
                DelayForLCDEnable();
                E_PIN = 1;              // Clock out other nibble
                DelayForLCDEnable();
                E_PIN = 0;
                RW_PIN = 0;             // Reset control line
                return TRUE;               // Return TRUE
        }
        else                            // Busy bit is low
        {
                E_PIN = 0;              // Reset clock line
                DelayForLCDEnable();
                E_PIN = 1;              // Clock out other nibble
                DelayForLCDEnable();
                E_PIN = 0;
                RW_PIN = 0;             // Reset control line
                return FALSE;               // Return FALSE
        }

}

/********************************************************************
* Function Name:  SetCGRamAddr
* Return Value:   void
* Parameters:     CGaddr: character generator ram address
* Description:    This routine sets the character generator address
*                 of the Hitachi HD44780 LCD controller. The
*				  user must check to see if the LCD controller
*				  is busy before calling this routine
********************************************************************/
void SetCGRamAddr(unsigned char CGaddr)
{
        TRIS_DATA_PORT &= 0xf0;			// Set data bits to outputs
        DATA_PORT &= 0xf0;				// and write upper nibble
        DATA_PORT |= (((CGaddr | LCD_CGRADDR)>>4) & 0x0f);

        RW_PIN = 0;						// Set control signals
        RS_PIN = 0;
        mPulseXLCDEnable();				// Clock the cmd in
        DATA_PORT &= 0xf0;				// Write lower nibble
        DATA_PORT |= (CGaddr&0x0f);

        DelayForLCDEnable();
        mPulseXLCDEnable();				// Clock the cmd in

        TRIS_DATA_PORT |= 0x0f;			// Restore data bits to inputs

        return;
}


/********************************************************************
* Function Name:  SetDDRamAddr
* Return Value:   void
* Parameters:     CGaddr: display data address
* Description:    This routine sets the display data address
*                 of the Hitachi HD44780 LCD controller. The
*                 user must check to see if the LCD controller
*                 is busy before calling this routine.
********************************************************************/
void SetDDRamAddr(unsigned char DDaddr)
{
        TRIS_DATA_PORT &= 0xf0;			// Set data bits to outputs
        DATA_PORT &= 0xf0;              // Write upper nibble of address
        DATA_PORT |= (((DDaddr | LCD_DATADDR)>>4) & 0x0f);

        RW_PIN = 0;                     // Set control bits
        RS_PIN = 0;
		mPulseXLCDEnable();				// Clock the cmd in
        DATA_PORT &= 0xf0;              // Write lower nibble of address
        DATA_PORT |= (DDaddr&0x0f);

        DelayForLCDEnable();
        mPulseXLCDEnable();				// Clock the cmd in

        TRIS_DATA_PORT |= 0x0f;         // Restore data bits to inputs

        return;
}


/********************************************************************
*       Function Name:  WriteCmdXLCD                                *
*       Return Value:   void                                        *
*       Parameters:     cmd: command to send to LCD                 *
*       Description:    This routine writes a command to the Hitachi*
*                       HD44780 LCD controller. The user must check *
*                       to see if the LCD controller is busy before *
*                       calling this routine.                       *
********************************************************************/
void WriteCmdXLCD(unsigned char cmd)
{
        TRIS_DATA_PORT &= 0xf0;		// Set data bits to outputs
        DATA_PORT &= 0xf0;
        DATA_PORT |= (cmd>>4)&0x0f;	// Send hi nibble

        RW_PIN = 0;                 // Set control signals for command
        RS_PIN = 0;

    	mPulseXLCDEnable();			// Clock command in
                                  
        DATA_PORT &= 0xf0;
        DATA_PORT |= cmd&0x0f;		// Now the low nibble

        DelayForLCDEnable();
        mPulseXLCDEnable();

        TRIS_DATA_PORT |= 0x0f;		// Restore data bits to inputs
        return;
}

/********************************************************************
*       Function Name:  WriteDataXLCD                               *
*       Return Value:   void                                        *
*       Parameters:     data: data byte to be written to LCD        *
*       Description:    This routine writes a data byte to the      *
*                       Hitachi HD44780 LCD controller. The user    *
*                       must check to see if the LCD controller is  *
*                       busy before calling this routine. The data  *
*                       is written to the character generator RAM or*
*                       the display data RAM depending on what the  *
*                       previous SetxxRamAddr routine was called.   *
********************************************************************/
void WriteDataXLCD(char data)
{
        TRIS_DATA_PORT &= 0xf0;			// Set data bits to outputs
        DATA_PORT &= 0xf0;				// Write high nibble
        DATA_PORT |= ((data>>4)&0x0f);

        RS_PIN = 1;                     // Set control bits
        RW_PIN = 0;
        
    	mPulseXLCDEnable();				// Clock data in
                                  
        DATA_PORT &= 0xf0;				// Write low nibble
        DATA_PORT |= (data&0x0f);

        DelayForLCDEnable();
        mPulseXLCDEnable();

        TRIS_DATA_PORT |= 0x0f;			// Restore data bits to inputs
        return;
}

