/******************************************************************************
 *
 *  Filename:       mxg_xlcd.h
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 *
 *
 *****************************************************************************/
 
/******************************************************************************
 * mistralXG LCD utility routines (adapted from the Microchip library to avoid
 * blocking while writing to the LCD).
 *****************************************************************************/
#ifndef __MXG_XLCD_H
#define __MXG_XLCD_H

/* PIC18 XLCD peripheral routines.
 *
 * Author          Date       Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Steve Russell   04Mar08    Adapted from Microchip library (avoiding blocking)
 *
 */

// DATA_PORT defines the port to which the LCD data lines are connected 
// - uses 4-bit mode on lower nibble of the port 
#define DATA_PORT      PORTA
#define TRIS_DATA_PORT TRISA

// CTRL_PORT defines the port pins to which the control lines are connected.
#define RW_PIN   PORTAbits.RA5   /* PORT for RW */
#define TRIS_RW  DDRAbits.RA5    /* TRIS for RW */
#define RS_PIN   PORTAbits.RA4   /* PORT for RS */
#define TRIS_RS  DDRAbits.RA4    /* TRIS for RS */
#define E_PIN    PORTCbits.RC0   /* PORT for E  */
#define TRIS_E   DDRCbits.RC0    /* TRIS for E  */

// LCD Command defines
#define LCD_CLEAR	0b00000001
#define LCD_CGRADDR	0b01000000
#define LCD_DATADDR	0b10000000

// CG characters
#define NO_MIDI     0x0c        /* Character used for MIDI activity display */

// Display ON/OFF Control defines
#define DON         0b00001111  /* Display on      */
#define DOFF        0b00001011  /* Display off     */
#define CURSOR_ON   0b00001111  /* Cursor on       */
#define CURSOR_OFF  0b00001101  /* Cursor off      */
#define BLINK_ON    0b00001111  /* Cursor Blink    */
#define BLINK_OFF   0b00001110  /* Cursor No Blink */

// Cursor or Display Shift defines
#define SHIFT_CUR_LEFT    0b00010011  /* Cursor shifts to the left   */
#define SHIFT_CUR_RIGHT   0b00010111  /* Cursor shifts to the right  */
#define SHIFT_DISP_LEFT   0b00011011  /* Display shifts to the left  */
#define SHIFT_DISP_RIGHT  0b00011111  /* Display shifts to the right */

/* Function Set defines */
#define FOUR_BIT   0b00101111  /* 4-bit Interface               */
#define LINES_5X7  0b00111011  /* 5x7 characters, multiple line */

#define MEM_MODEL near  /* Change this to near for small memory model */

/* Useful LCD macros
 * 
 * E must be high for at least 230ns and have a cycle time no less than
 * 500ns. Delaying for four instruction cycles will achieve this (48MHz clock)
 *
 * mPulseXLCDEnable() drives the E pin high for the required time. At 48MHz
 * four instruction cycles is sufficient, so call/return to an empty subroutine
 * DelayForLCDEnable() is all that is needed.Depending on what follows, you 
 * may need another call to DelayForLCDEnable() to ensure E is low for long
 * enough to ensure the correct cycle time.
 */

#define mPulseXLCDEnable()  E_PIN = 1;DelayForLCDEnable();E_PIN = 0;		

/* OpenXLCD
 * Configures I/O pins for external LCD
 */
void OpenXLCD(unsigned char);

/* BusyXLCD
 * Returns the busy status of the LCD
 */
BOOL BusyXLCD(void);

/* WriteCmdXLCD
 * Writes a command to the LCD
 */
void WriteCmdXLCD(unsigned char);

/* WriteDataXLCD
 * Writes a data byte to the LCD
 */
void WriteDataXLCD(char);

/* SetDDRamAddr
 * Set the LCD data address
 */
void SetDDRamAddr(unsigned char);

/* SetCGRamAddr
 * Set the LCD character generator address
 */
void SetCGRamAddr(unsigned char);

/* User defines these routines according to the oscillator frequency */
extern void DelayForLCDEnable(void);

#endif
