/******************************************************************************
 *
 *  Filename:       user.c
 *  Project:        mistralXG Wind synthesiser
 *  Date:           12Apr08
 *  File Version:   01.00
 *
 *
 *  Author:     Steve Russell (pic.projects@ntlworld.com)
 *  Company:    Grapevyne
 *
 *  Copyright:  (C) Steve Russell 2008
 *
 *  This code is available at no charge to individuals for personal, non-profit,
 *  use, and may be used as is or modifed as desired. Further distribution of
 *  this code or a derivative work is allowed as long as this copyright notice
 *  is not removed or modified. Changes to the original code should be clearly
 *  identified.
 *
 *  Any commercial use of this code, or a derivative of it, may only be made
 *  with the written agreement of the copyright holder. "Commercial" in this
 *  context means any activity that requires a user to pay in money or kind
 *  in order to make use of such code.
 *
 *  This file is provided "as is". No claims of suitability for any particular
 *  purpose are made. No liability for any damages (of any sort) resulting from
 *  its use are accepted, no matter how they may arise.
 *
 ******************************************************************************
 * Change log
 * ----------
 * 08Apr08 - Successfully talking to the LCD without blocking
 * 12Apr08 - EUSART interrupt and data handling to buffers complete but
 *           simulator doesn't work. Pushed megabytes through RX/TX without
 *           problem. Raised a simulator bug with Microchip.
 * 23Apr08 - Microchip confirm bug in simulator
 * 26Apr08 - MIDI data sent from the PIC to my PC!
 * 03May08 - MIDI data successfully sent in both directions simultaneously.
 *           MIDI songs played PC>USB>PIC>USB>PC.
 * 20May08 - LCD/button user interface functional. Not yet controlling MIDI
 *           streams, but can flip through different displays.
 * 01Jul08 - Flipping outputs to control streams now. Changed to bit array for 
 *           options so that we can store easily in EEPROM. User options now
 *           stored in EEPROM so that previous settings are restored when power
 *           is applied.
 * 16Jul08 - USB connect/disconnect now shown on display and used to switch
 *           MIDI data stream to previously selected external source if USB is
 *           selected when the cable is pulled.
 * 31Jul08 - Stop EUSART Rx/Tx when USB is disconnected
 * 06Aug08 - Added MIDI IN and MIDI OUT monitoring to the Home display
 * 09Aug08 - Use USBActive to prevent MIDI being sent when not connected
 * 08Sep08 - Added reset all error counters code (press Set while at an error
 *           counter that is already zero).
 * 09Sep08 - Some tidying up and checked usage of all variables modified by
 *           the interrupt handler, making changes to try to ensure no race
 *           conditions remain.
 * 25Sep08 - Save and restore brightness from EEPROM
 *
 *****************************************************************************/

/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include "system\typedefs.h"
#include "system\usb\usb.h"
#include "system\interrupt\interrupt.h"

#include "io_cfg.h"             // I/O pin mapping
#include "user\user.h"
#include "user\mxg_usart.h"
#include "user\mxg_musm.h"
#include "user\mxg_umsm.h"
#include "user\mxg_error.h"
#include "user\mxg_msgs.h"
#include "user\mxg_xlcd.h"
#include "user\mxg_iface.h"
#include "user\mxg_eeprom.h"

/** P R I V A T E  P R O T O T Y P E S ***************************************/
void ServiceRequests(void);
void setOptions(void);

/** V A R I A B L E S ********************************************************/
// User switches - initialised by setOptions()
// These used to be BOOLs (we could spare the space!) but were changed to
// bits to simplify their storage in EEPROM (see user.h)
BYTE    userOptions;
/*
BOOL	WX5Sel;			// TRUE: WX5 data, FALSE: Ext MIDI IN data
BOOL	EnIntMIDI;		// TRUE: Data from PC USB, FALSE: Data from WX5/Ext
BOOL	EnMIDIThru;		// TRUE: Enable MIDI Thru from WX5/Ext
BOOL	EnMIDIOut;		// TRUE: Enables PC USB MIDI output
BOOL	EnMIDIIn;		// TRUE: Allow Ext/WX5 MIDI through to USB
BOOL	FilterMIDIIn;	// TRUE: Remove invalid data from input stream to USB
BOOL	EnRunStatus;	// TRUE: Enable running status on output stream from USB
*/

// Flag to let us detect when the USB port has just been connected
BOOL USBActive;

/** D E C L A R A T I O N S **************************************************/
#pragma code

/******************************************************************************
 * Function:        void UserInit(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        Initialise peripherals and flags etc. used by the app
 * Note:            IPEN (RCON<7> is 0 at POR, so no interrupt priority
 *****************************************************************************/
void UserInit(void)
{	
	mInitIO();                  // Set up the IO pins
    initEEPROM();               // Get the EEPROM ready
    clearErrors();              // Clear all the error flags & counters
	setOptions();               // Set user options from their saved settings

    // Init user interface
    initInterface();

    // Assume the USB port is inactive to start with
    USBActive = FALSE;          // Clear flag
    bIntMIDISel = 0;            // Switch to external (WX5/MIDI) MIDI
    USBFLAG = USBDISABLED;                  // Reflect on display
    MIDIFLAG = WX5Sel ? WX5IN : MIDIIN;     // Show the relevant source

    // Init the USART to receive data for monitoring even if USB is inactive
    initUSART();
    // Init MIDI state machines
    initMusm();         // Set up the MIDI receive state machine
    initUmsm();         // Set up the MIDI transmit state machine

    // Make sure Tx interrupts are off until we have something to send
	mDisableTxInt();

	// Set up Timer 0 - internal clock source, 8-bit, 1/64 prescale
	T0CON=0b11010101;
    // Initialise system timers
    // Interrupts not running yet, so no need to be careful
    // NOTE: display timers set in setDisplay()
    mEnableTimerInt();					// Enable timer 0 interrupts
	mClearTimerInt();					// Clear timer 0 flag
	TMR0L=INTERVAL;						// Set the timer interval (1ms)
	// Off we go... enable all interrupts
	INTCONbits.PEIE=1;					// First the peripherals
	INTCONbits.GIE=1;					// ... now everything
	// Set up the display (needs timer interrupt running)
    OpenXLCD(FOUR_BIT & LINES_5X7);
}//end UserInit


/******************************************************************************
 * Function:        void ProcessIO(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        
 * Note:            None
 *****************************************************************************/
void ProcessIO(void)
{
	// Ok, let's process the MIDI IN data first, limiting ourselves to only
    // one byte each time round the loop. This won't be a problem as a byte
    // takes 320us to arrive at MIDI speeds and less than 10us to process.
    // By measurement, the ProcessIO loop only takes around 12us typically,
    // and not much longer worst case. mRxDataRdy() is interrupt safe.
    if(mRxDataRdy()) {
        processNextByte();
    }

    if(bChange){            // Has a button state changed?
        switch(savedSwStatus) {
            case bSelect:
                doSelect();
                break;
            case bSet:          
                doSet();
                break;
            case bBoth:
            case bNone:
            default:
                break;
        }
        bChange=FALSE;
    }

    // Check whether the EEPROM has started a write and, if so,
    // whether it is complete or not
    if(writingEEPROM && PIR2bits.EEIF) {
        EECON1bits.WREN = 0;        // Yes - disable writes
        PIR2bits.EEIF = 0;          // ... and clear the interrupt
        writingEEPROM=FALSE;        // ... and the flag
    }

    // Check if the user changed a setting... if so, write it to EEPROM
    if(optsChange && !writingEEPROM){
        optsChange = FALSE;         // Clear the change bit
        writeEEPROM(userOptions._byte, SAVED_OPTS);
    }

    // Check if the LCD brightness has changed... if so, write it to EEPROM
    if(brightChange && !writingEEPROM){
        brightChange = FALSE;         // Clear the change bit
        writeEEPROM(brightness, SAVED_BRITE);
    }

	// Checking dispTimeout is not atomic, but we can avoid disabling
	// interrupts by checking the high byte and then the low byte in that
	// order (because dispTimeout is decremented).
	if(!dispTimeout.HighB._byte && !dispTimeout.LowB._byte) {
        setDisplay(dHOME);
    }
    if(!displayTmr){
        refreshDisplay();
    }
    if (WritesToLCDPending) {	        // We're writing to the display
        if(!BusyXLCD()){
            nextChar();   // Send the next char if not busy
        }
    }


    // User Application USB tasks
    if((usb_device_state < CONFIGURED_STATE)||(UCONbits.SUSPND==1)) {
        // USB either disconnected or suspended so not in use...
        // Do some housekeeping if the USB port has just gone out of service
        if(USBActive) {
            // USBActive lets us keep receiving MIDI to provide monitoring
            // and prevent MIDI buffer overflow when not connected
            USBActive = FALSE;          // Clear flag so we only do this once
            bIntMIDISel = 0;            // Switch to external (WX5/MIDI) MIDI
            USBFLAG = USBDISABLED;                  // Reflect on display
            MIDIFLAG = WX5Sel ? WX5IN : MIDIIN;     // Show the relevant source
        }
        // ... and return
        return;
    }
    else {      // USB is active
        if(!USBActive) {        // If USB has just started up
            USBActive = TRUE;   // Set flag so we can tell when USB stops
            clearErrors();		// Clear all the error flags & counters
            initMusm();         // Set up the MIDI receive state machine
            initUmsm();         // Set up the MIDI transmit state machine
            initUSART();        // Reset the EUSART ready for USB activity
            USBFLAG = USBENABLED;       // Reflect on display
            if (EnIntMIDI) {            // If flag is set...
                bIntMIDISel = 1;        // ...revert to USB stream
            }    
        }

        // Update the display to reflect any changes in MIDI source. This
        // has to be done here instead of in doSet() (mxg_iface.c) because we
        // want to change the source when USB is connected/disconnected.
        if (EnIntMIDI) {            // If flag is set...
            MIDIFLAG = USBIN;       // ...show USB as source
        }
        else {                      // Otherwise show the selected ext source
            MIDIFLAG = WX5Sel ? WX5IN : MIDIIN;
        }
        ServiceRequests();
    }
}//end ProcessIO

/******************************************************************************
 * This routine handles the MIDI I/O between the PC USB port and the MIDI
 * IN and MIDI OUT ports on the board by calling the appropriate routines in
 * the Microchip stack (essentially the same as provided... just text 
 * modifications to reference the MIDI class)
 *****************************************************************************/
void ServiceRequests(void)
{
    // If there's some data to send and the USB SIE isn't already busy, send it
    if(UMIptr.bytes.lo&0x3f && !mUSBMSTxIsBusy()) {
        USBMSWrite(USBMidiIn, UMIptr.bytes.lo&0x3f);
        UMIptr.bytes.lo=0xc0;       // Empty the buffer
    }

    // Make sure that we have enough space in the transmit buffer to accommodate
    // 48 bytes of MIDI data (the most that can be sent in a single transfer to
    // the USB endpoint (64 bytes, or 16 USB-MIDI packets). We don't have to
    // worry about interrupts for this calculation as the buffer free space 
    // will only get bigger.
    if((((txbend.bytes.lo|0x80)-txbstart.bytes.lo)&0x7F)<=(127-48)){
        // Now see if we've received any data from the USB port, and send the 
        // corresponding MIDI data out to the DB50XG and/or the external 
        // MIDI OUT. Test for SIE busy is handled inside USBMSRead()
        // ...returns 0 if no data
        if(USBMSRead(USBMidiOut, USBMS_EP_SIZE)){
            processNextBuffer();        // See mxg_umsm.c
        }
    }
}//end ServiceRequests


/******************************************************************************
 * Set the various user options, using the previous settings read from EEPROM.
 * The first time the MCU is powered on these will be set to the default EEPROM
 * state. User settings will be maintained from then onwards.
 *****************************************************************************/
void setOptions()
{
    // Read the options byte stored in EEPROM
    userOptions._byte = readEEPROM(SAVED_OPTS);

    // Now set the display and switches appropriately
    // The MIDI source (WX5, MIDI, USB) for data going to the DB50XG is also
    // handled in the main loop to deal with the USB cable being unplugged.
    if(WX5Sel) {
        bWX5Sel = 1;
        MIDIFLAG = WX5IN;
    }
    else {
        bWX5Sel = 0;
        MIDIFLAG = MIDIIN;
    }

    // If USB is selected it overrides the external sources
    if(EnIntMIDI){
        bIntMIDISel = 1;
        MIDIFLAG = USBIN;
    }
    else{       // Leave the appropriate external source displayed
        bIntMIDISel = 0;
    }

    // Set the MIDI Thru flag
    if(EnMIDIThru){
        THRUFLAG = EnThruTRUE;
        bEnMIDIThru = 1;
    }
    else{
        THRUFLAG = EnThruFALSE;
        bEnMIDIThru = 0;
    }

    // ...and the MIDI Out flag
    if(EnMIDIOut){
        OUTFLAG = EnOutTRUE;
        bEnMIDIOut = 1;
    }
    else{
        OUTFLAG = EnOutFALSE;
        bEnMIDIOut = 0;
    }

    // Finally, the Running status and Bad MIDI filter flags
    RSFLAG = EnRunStatus?EnRSTRUE:EnRSFALSE;
    FILTFLAG = FilterMIDIIn?EnFiltTRUE:EnFiltFALSE;
}
/** EOF user.c ***************************************************************/
